@@ -5,6 +5,7 @@ Created on Wed Jan 20 10:43:15 2021
 
 @author: mike
 """
+import pandas as pd
 from io import BytesIO
 import xarray as xr
 import os
@@ -14,12 +15,14 @@ import requests
 import orjson
 import copy
 import multiprocessing as mp
-# from cdo import Cdo
 import pathlib
 from tethysts.utils import get_object_s3, url_stream_to_file, read_pkl_zstd, b2_public_key_pattern
 import gzip
 import zstandard as zstd
 import subprocess
+import glob
+import rioxarray as rxr
+import concurrent.futures
 # mp.set_start_method('spawn')
 
 ##########################################
@@ -67,7 +70,7 @@ def query_esri_mapserver(base_url, layer_id, out_fields=None, where=None):
         raise ValueError('where must either be None or a string')
 
     ## Run the queries
-    resp = requests.get(url, params=params)
+    resp = requests.get(url, params=params, timeout=300)
 
     if not resp.ok:
         raise ValueError(resp.raise_for_status())
@@ -89,7 +92,7 @@ def query_esri_mapserver(base_url, layer_id, out_fields=None, where=None):
         else:
             params.update({'where': 'objectid > ' + str(last_id)})
 
-        resp = requests.get(url, params=params)
+        resp = requests.get(url, params=params, timeout=300)
 
         if not resp.ok:
             raise ValueError(resp.raise_for_status())
@@ -167,18 +170,49 @@ def convert_gribs_remove(source_paths, max_workers=4):
     return results
 
 
-# def copy_file(source_path, dest_path, buffer_size=10485760):
-#     dest_path1 = pathlib.Path(dest_path)
-#     if dest_path1.is_dir():
-#         source_path1 = pathlib.Path(source_path)
-#         file_name = source_path1.name
-#         dest_path2 = str(dest_path1.joinpath(file_name))
-#     else:
-#         dest_path2 = dest_path
-#     with open(source_path, 'rb') as s_file, open(dest_path2, 'wb') as d_file:
-#         shutil.copyfileobj(s_file, d_file, buffer_size)
+def convert_geotiff_to_nc(source_path, time, height, x_name='x', y_name='y', band=1):
+    """
 
-#     return dest_path2
+    """
+    time1 = pd.Timestamp(time)
+
+    xr1 = rxr.open_rasterio(source_path)
+    xr1 = xr1.rename({x_name: 'lon', y_name: 'lat'}).sel(band=band).drop('band')
+    xr1 = xr1.assign_coords(height=height).expand_dims('height', axis=2)
+    xr1 = xr1.assign_coords(time=time1).expand_dims('time')
+    xr1.name = 'raster'
+
+    attrs = xr1.attrs.copy()
+
+    encoding = {'dtype': xr1.encoding['rasterio_dtype'], 'scale_factor': attrs['scale_factor'], 'add_offset': attrs['add_offset'], '_FillValue': -9999}
+
+    xr1.encoding = encoding
+    xr1.attrs = {}
+
+    dest_path = os.path.splitext(source_path)[0]
+    dest_path = dest_path + '.nc'
+
+    xr1.to_netcdf(dest_path)
+
+    os.remove(source_path)
+
+    return dest_path
+
+
+def convert_geotiffs_to_nc(source_paths, time, height, x_name='x', y_name='y', band=1, max_workers=4):
+    """
+
+    """
+    with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+        futures = []
+        for source_path in source_paths:
+            f = executor.submit(convert_geotiff_to_nc, source_path, time, height, x_name, y_name, band)
+            futures.append(f)
+        runs = concurrent.futures.wait(futures)
+
+    new_paths = [r.result() for r in runs[0]]
+
+    return new_paths
 
 
 def copy_file(source_path, dest_path, compression=None, chunk_size=524288):
@@ -241,6 +275,24 @@ def mergetime_nc_remove(source_paths, by):
     return dest_path
 
 
+def gdal_merge(source_paths, dest_path, compression='ZSTD'):
+    """
+    Uses gdal_merge.py to merge many rasters together using a defined compression to a GeoTiff. The compression must be one of the COMPRESS options for GeoTiffs:
+        https://gdal.org/drivers/raster/gtiff.html#raster-gtiff
+    """
+    f2 = source_paths.copy()
+    f2.sort()
+
+    source_paths_str = ' '.join(f2)
+
+    if not dest_path.endswith('.tif'):
+        raise ValueError('The dest_path must be a file path with a .tif extension.')
+
+    cmd_str = 'gdal_merge.py -co COMPRESS={compression} -o {out_file} {in_files}'.format(in_files=source_paths_str, out_file=dest_path, compression=compression)
+
+    _ = subprocess.run(cmd_str, shell=True, check=True)
+
+
 def copy_s3_file(source_path, dest_path, bucket, s3=None, connection_config=None, public_url=None, compression=None, chunk_size=524288):
     """
 
@@ -261,7 +313,7 @@ def copy_s3_file(source_path, dest_path, bucket, s3=None, connection_config=None
             raise gzip.BadGzipFile('Too many crc32 errors for: ' + source_path)
         try:
             if isinstance(public_url, str):
-                url1 = b2_public_key_pattern.format(base_url=public_url, bucket=bucket, obj_key=source_path)
+                url1 = b2_public_key_pattern.format(base_url=public_url.rstrip('/'), bucket=bucket, obj_key=source_path)
                 file_path3 = url_stream_to_file(url1, file_path2, compression=compression, chunk_size=chunk_size)
             else:
                 url1 = get_object_s3(source_path, connection_config=connection_config, bucket=bucket, counter=5)
