@@ -83,6 +83,9 @@ class Raster(tu.titan.Titan):
     """
     ## Initial import and assignment function
     def import_raster(self, source_paths, time, height, x_name='x', y_name='y', band=1):
+        """
+
+        """
         tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
 
         f_dict, max_f = parse_images(source_paths)
@@ -104,7 +107,7 @@ class Raster(tu.titan.Titan):
         return crs.to_proj4()
 
 
-    def merge_rasters(self, remove_source_images=False):
+    def spatial_merge_rasters(self, remove_source_images=False):
         """
 
         """
