@@ -7,25 +7,16 @@ Created on Sat May  1 13:30:43 2021
 """
 import os
 import numpy as np
-import zstandard as zstd
 import pandas as pd
-import xarray as xr
 import orjson
-from tethys_data_models.dataset import Dataset, Station
-# from data_models import Geometry, Dataset, DatasetBase, S3ObjectKey
 import tethys_data_models as tdm
-from tethysts.utils import get_object_s3, s3_client, read_json_zstd, read_pkl_zstd, chunk_filters
-from tethys_utils.misc import make_run_date_key, write_json_zstd, path_date_parser
-# from misc import make_run_date_key, write_json_zstd, write_pkl_zstd
-from tethys_utils.processing import stats_for_dataset_metadata
-# from processing import compare_xrs, process_station_summ, stats_for_dataset_metadata
-from datetime import date, datetime
+from tethysts.utils import get_object_s3, read_json_zstd, s3_client
+from tethys_utils.misc import make_run_date_key, path_date_parser, write_json_zstd
 from tethysts import Tethys
 from botocore import exceptions as bc_exceptions
 from time import sleep
 import pathlib
 import concurrent.futures
-import multiprocessing as mp
 import copy
 
 ############################################
@@ -61,6 +52,8 @@ def put_object_s3(s3, bucket, key, obj, metadata, content_type, retries=5):
             print('...trying again...')
             sleep(5)
 
+    obj2.update({'key': key, 'bucket': bucket})
+
     return obj2
 
 
@@ -106,7 +99,7 @@ def copy_object_s3(s3, source_bucket, dest_bucket, source_key, dest_key, retries
 
 def multi_copy_object_s3(s3, source_bucket, dest_bucket, source_dest_keys, retries=5, threads=30):
     """
-    Same as the copy_object_s3 except with multi threading. the input source_dest_keys must be a list of dictionaries with keys named source_key and dest_key.
+    Same as the copy_object_s3 except with multi threading. The input source_dest_keys must be a list of dictionaries with keys named source_key and dest_key.
     """
     keys = copy.deepcopy(source_dest_keys)
 
@@ -260,71 +253,75 @@ def list_object_versions_s3(s3_client, bucket, prefix, start_after='', delimiter
     return f_df1
 
 
-def put_dataset(dataset, bucket, s3=None, connection_config=None, public_url=None, system_version=4):
-    """
+# def put_dataset(dataset, bucket, s3=None, connection_config=None, public_url=None, system_version=4):
+#     """
 
-    """
-    run_date_key = make_run_date_key()
+#     """
+#     run_date_key = make_run_date_key()
 
-    dataset_id = dataset['dataset_id']
+#     dataset_id = dataset['dataset_id']
 
-    ## Get the stations agg
-    stns_key = tdm.utils.key_patterns[system_version]['stations'].format(dataset_id=dataset_id)
+#     ## Get the stations agg
+#     stns_key = tdm.utils.key_patterns[system_version]['stations'].format(dataset_id=dataset_id)
 
-    obj1 = get_object_s3(stns_key, bucket, s3, connection_config, public_url)
-    stns = read_json_zstd(obj1)
+#     obj1 = get_object_s3(stns_key, bucket, s3, connection_config, public_url)
+#     stns = read_json_zstd(obj1)
 
-    ## generate stats for the dataset metadata
-    ds_stats = stats_for_dataset_metadata(stns)
+#     ## generate stats for the dataset metadata
+#     ds_stats = stats_for_dataset_metadata(stns)
 
-    if dataset['result_type'] == 'time_series':
-        if 'spatial_resolution' in ds_stats:
-            ds_stats.pop('spatial_resolution')
+#     if dataset['result_type'] == 'time_series':
+#         if 'spatial_resolution' in ds_stats:
+#             ds_stats.pop('spatial_resolution')
 
-    dataset.update(ds_stats)
+#     dataset.update(ds_stats)
 
-    ## Add version number
-    dataset.update({'system_version': system_version})
+#     ## Add version number
+#     dataset.update({'system_version': system_version})
 
-    ## Check and create dataset metadata
-    ds4 = tdm.dataset.Dataset(**dataset)
+#     ## Check and create dataset metadata
+#     ds4 = tdm.dataset.Dataset(**dataset)
 
-    ds5 = orjson.loads(ds4.json(exclude_none=True))
+#     ds5 = orjson.loads(ds4.json(exclude_none=True))
 
-    ## Write the object
-    ds_obj = write_json_zstd(ds5)
+#     ## Write the object
+#     ds_obj = write_json_zstd(ds5)
 
-    ds_key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
+#     ds_key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
 
-    _ = put_object_s3(s3, bucket, ds_key, ds_obj, {'run_date': run_date_key}, 'application/json')
+#     _ = put_object_s3(s3, bucket, ds_key, ds_obj, {'run_date': run_date_key}, 'application/json')
 
-    return ds5
+#     return ds5
 
 
-def get_remote_dataset(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, system_version=4):
+# def get_remote_dataset(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, system_version=4):
+#     """
+
+#     """
+#     if isinstance(dataset_id, str):
+#         key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
+#     elif not isinstance(key, str):
+#         raise ValueError('dataset_id must be passed or key must be passed.')
+
+#     try:
+#         obj1 = get_object_s3(key, bucket, s3, connection_config, public_url)
+#         rem_ds = read_json_zstd(obj1)
+#     except:
+#         rem_ds = None
+
+#     return rem_ds
+
+
+def get_remote_stations(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, version_date=None, system_version=4):
     """
 
     """
     if isinstance(dataset_id, str):
-        key = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
-    elif not isinstance(key, str):
-        raise ValueError('dataset_id must be passed or key must be passed.')
-
-    try:
-        obj1 = get_object_s3(key, bucket, s3, connection_config, public_url)
-        rem_ds = read_json_zstd(obj1)
-    except:
-        rem_ds = None
-
-    return rem_ds
-
-
-def get_remote_station(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, station_id=None, key=None, system_version=4):
-    """
-
-    """
-    if isinstance(dataset_id, str) and isinstance(station_id, str):
-        key = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=dataset_id, station_id=station_id)
+        if isinstance(version_date, (str, pd.Timestamp)):
+            vd = make_run_date_key(version_date)
+            key = tdm.utils.key_patterns[system_version]['stations'].format(dataset_id=dataset_id, version_date=vd)
+        else:
+            key = tdm.utils.key_patterns[system_version]['latest_stations'].format(dataset_id=dataset_id)
     elif not isinstance(key, str):
         raise ValueError('dataset_id and station_id must be passed or key must be passed.')
 
@@ -337,12 +334,16 @@ def get_remote_station(bucket, s3=None, connection_config=None, public_url=None,
     return rem_stn
 
 
-def get_remote_results_chunks(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, system_version=4):
+def get_remote_results_chunks(bucket, s3=None, connection_config=None, public_url=None, dataset_id=None, key=None, version_date=None, system_version=4):
     """
 
     """
     if isinstance(dataset_id, str):
-        key = tdm.utils.key_patterns[system_version]['results_chunks'].format(dataset_id=dataset_id)
+        if isinstance(version_date, (str, pd.Timestamp)):
+            vd = make_run_date_key(version_date)
+            key = tdm.utils.key_patterns[system_version]['results_chunks'].format(dataset_id=dataset_id, version_date=vd)
+        else:
+            key = tdm.utils.key_patterns[system_version]['latest_results_chunks'].format(dataset_id=dataset_id)
     elif not isinstance(key, str):
         raise ValueError('dataset_id must be passed or key must be passed.')
 
@@ -355,141 +356,150 @@ def get_remote_results_chunks(bucket, s3=None, connection_config=None, public_ur
     return rem_ds
 
 
-def put_remote_agg_stations(dataset_id, bucket, s3, public_url=None, threads=60, system_version=4):
-    """
+# def put_remote_stations(dataset_id, version_date, bucket, s3, public_url=None, system_version=4):
+#     """
+#     Needs to be updated to create the version specific stations and results chunks files.
+#     """
+#     stn_key = tdm.utils.key_patterns[system_version]['stations']
+#     vd = make_run_date_key(version_date)
+#     stn_prefix = stn_key.split('{station_id}')[0].format(dataset_id=dataset_id, verion_date=vd)
 
-    """
-    stn_key = tdm.utils.key_patterns[system_version]['station']
-    stn_prefix = stn_key.split('{station_id}')[0].format(dataset_id=dataset_id)
+#     if system_version >= 4:
+#         list1 = list_objects_s3(s3, bucket, stn_prefix, delimiter='/')
+#     else:
+#         raise ValueError('Wrong object structure version.')
 
-    if system_version == 2:
-        list1 = list_objects_s3(s3, bucket, stn_prefix)
-    elif system_version >= 3:
-        list1 = list_objects_s3(s3, bucket, stn_prefix, delimiter='/')
-    else:
-        raise ValueError('Wrong object structure version.')
+#     list2 = list1[list1.Key.str.contains('station.json.zst')].copy()
 
-    list2 = list1[list1.Key.str.contains('station.json.zst')].copy()
+#     if not list2.empty:
 
-    if not list2.empty:
+#         ## Get all of the result chunks from the individual stations
+#         with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+#             futures = []
 
-        ## Get all of the result chunks from the individual stations
-        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
-            futures = []
+#             for key in list2['Key'].tolist():
+#                 f = executor.submit(get_remote_station, s3=s3, bucket=bucket, key=key, public_url=public_url, system_version=system_version)
+#                 futures.append(f)
 
-            for key in list2['Key'].tolist():
-                f = executor.submit(get_remote_station, s3=s3, bucket=bucket, key=key, public_url=public_url, system_version=system_version)
-                futures.append(f)
+#             runs = concurrent.futures.wait(futures)
 
-            runs = concurrent.futures.wait(futures)
+#         stn_list = [r.result() for r in runs[0]]
 
-        stn_list = [r.result() for r in runs[0]]
+#         ## Remove the results chunks from the stations
+#         rc_list = []
+#         _ = [rc_list.extend(s.pop('results_chunks')) for s in stn_list]
 
-        ## Remove the results chunks from the stations
-        _ = [s.pop('results_chunks') for s in stn_list]
+#         ## Save files
+#         # Stations
+#         stns_obj = write_json_zstd(stn_list)
 
-        ## Save files
-        # Stations
-        stns_obj = write_json_zstd(stn_list)
+#         run_date_key = make_run_date_key()
 
-        run_date_key = make_run_date_key()
+#         agg_stn_key = tdm.utils.key_patterns[system_version]['stations']
 
-        agg_stn_key = tdm.utils.key_patterns[system_version]['stations']
+#         stns_key = agg_stn_key.format(dataset_id=dataset_id, verion_date=vd)
 
-        stns_key = agg_stn_key.format(dataset_id=dataset_id)
+#         _ = put_object_s3(s3, bucket, stns_key, stns_obj, {'run_date': run_date_key}, 'application/json')
 
-        _ = put_object_s3(s3, bucket, stns_key, stns_obj, {'run_date': run_date_key}, 'application/json')
+#         # Results chunks
+#         rc_obj = write_json_zstd(rc_list)
 
-        return stn_list
-    else:
-        print('There are no stations files in the database.')
-        return None
+#         agg_rc_key = tdm.utils.key_patterns[system_version]['results_chunks']
+
+#         rc_key = agg_rc_key.format(dataset_id=dataset_id, verion_date=vd)
+
+#         _ = put_object_s3(s3, bucket, rc_key, rc_obj, {'run_date': run_date_key}, 'application/json')
+
+#         return stn_list
+#     else:
+#         print('There are no stations files in the database.')
+#         return None
 
 
-def put_remote_agg_datasets(bucket, s3, public_url=None, threads=30, system_version=4):
-    """
+# def put_remote_agg_datasets(bucket, s3, public_url=None, threads=30, system_version=4):
+#     """
 
-    """
-    ds_key = tdm.utils.key_patterns[system_version]['dataset']
+#     """
+#     ds_key = tdm.utils.key_patterns[system_version]['dataset']
 
-    ds_prefix = ds_key.split('{dataset_id}')[0]
+#     ds_prefix = ds_key.split('{dataset_id}')[0]
 
-    if system_version == 2:
-        list1 = list_objects_s3(s3, bucket, ds_prefix)
-    elif system_version >= 3:
-        list1 = list_objects_s3(s3, bucket, ds_prefix, delimiter='/')
-    else:
-        raise ValueError('Wrong object structure version.')
+#     if system_version == 2:
+#         list1 = list_objects_s3(s3, bucket, ds_prefix)
+#     elif system_version >= 3:
+#         list1 = list_objects_s3(s3, bucket, ds_prefix, delimiter='/')
+#     else:
+#         raise ValueError('Wrong object structure version.')
 
-    list2 = list1[list1.Key.str.contains('dataset.json.zst')].copy()
+#     list2 = list1[list1.Key.str.contains('dataset.json.zst')].copy()
 
-    if not list2.empty:
+#     if not list2.empty:
 
-        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
-            futures = []
+#         with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+#             futures = []
 
-            for key in list2['Key'].tolist():
-                f = executor.submit(get_remote_dataset, s3=s3, bucket=bucket, public_url=public_url, key=key, system_version=system_version)
-                futures.append(f)
+#             for key in list2['Key'].tolist():
+#                 f = executor.submit(get_remote_dataset, s3=s3, bucket=bucket, public_url=public_url, key=key, system_version=system_version)
+#                 futures.append(f)
 
-            runs = concurrent.futures.wait(futures)
+#             runs = concurrent.futures.wait(futures)
 
-        ds_list = [r.result() for r in runs[0]]
+#         ds_list = [r.result() for r in runs[0]]
 
-        dss_obj = write_json_zstd(ds_list)
+#         dss_obj = write_json_zstd(ds_list)
 
-        run_date_key = make_run_date_key()
-        dss_key = tdm.utils.key_patterns[system_version]['datasets']
+#         run_date_key = make_run_date_key()
+#         dss_key = tdm.utils.key_patterns[system_version]['datasets']
 
-        _ = put_object_s3(s3, bucket, dss_key, dss_obj, {'run_date': run_date_key}, 'application/json')
+#         _ = put_object_s3(s3, bucket, dss_key, dss_obj, {'run_date': run_date_key}, 'application/json')
 
-    return ds_list
+#     return ds_list
 
 
-def update_versions(version_data, bucket, connection_config, public_url=None, system_version=4):
-    """
+# def update_versions(version_data, bucket, connection_config=None, public_url=None, system_version=4):
+#     """
 
-    """
-    ## Check version_dict data model
-    version_dict1 = tdm.dataset.ResultVersion(**version_data).dict(exclude_none=True)
+#     """
+#     ## Check version_dict data model
+#     version_dict1 = tdm.dataset.ResultVersion(**version_data).dict(exclude_none=True)
 
-    ## Initial set up
-    s3 = s3_client(connection_config)
+#     ## Initial set up
+#     # s3 = s3_client(connection_config)
 
-    dataset_id = version_dict1['dataset_id']
+#     dataset_id = version_dict1['dataset_id']
 
-    ## Get the versions
-    rv_key = tdm.utils.key_patterns[system_version]['versions'].format(dataset_id=dataset_id)
+#     ## Get the versions
+#     rv_key = tdm.utils.key_patterns[system_version]['versions'].format(dataset_id=dataset_id)
 
-    v_obj1 = get_object_s3(rv_key, bucket, s3=s3, public_url=public_url, counter=1)
+#     v_obj1 = get_object_s3(rv_key, bucket, connection_config=connection_config, public_url=public_url, counter=5)
 
-    if v_obj1 is not None:
-        v2 = read_json_zstd(v_obj1)
+#     if v_obj1 is not None:
+#         v2 = read_json_zstd(v_obj1)
 
-        exists = any([version_dict1['version_date'].isoformat() == v['version_date'] for v in v2])
+#         exists = any([version_dict1['version_date'].isoformat() == v['version_date'] for v in v2])
 
-        if exists:
-            version_list = []
-            for v in v2:
-                if version_dict1['version_date'].isoformat() == v['version_date']:
-                    version_list.append(version_dict1)
-                else:
-                    version_list.append(v)
-        else:
-            version_list = v2.copy()
-            version_list.append(version_dict1)
-    else:
-        version_list = [version_dict1]
+#         if exists:
+#             version_list = []
+#             for v in v2:
+#                 if version_dict1['version_date'].isoformat() == v['version_date']:
+#                     version_list.append(version_dict1)
+#                 else:
+#                     version_list.append(v)
+#         else:
+#             version_list = v2.copy()
+#             version_list.append(version_dict1)
+#     else:
+#         version_list = [version_dict1]
 
-    ## Save files
-    run_date_key = make_run_date_key()
+#     ## Save files
+#     # run_date_key = make_run_date_key()
 
-    # versions
-    rv_obj = write_json_zstd(version_list)
+#     # # versions
+#     # rv_obj = write_json_zstd(version_list)
 
-    _ = put_object_s3(s3, bucket, rv_key, rv_obj, {'run_date': run_date_key}, 'application/json')
+#     # _ = put_object_s3(s3, bucket, rv_key, rv_obj, {'run_date': run_date_key}, 'application/json')
 
-    return version_list
+#     return version_list
 
 
 # def update_station_s3(stn_data, bucket, connection_config, public_url=None, system_version=4):
@@ -533,119 +543,125 @@ def update_versions(version_data, bucket, connection_config, public_url=None, sy
 #     return stn_dict2
 
 
-def put_result(s3, bucket, results_path, system_version=4):
-    """
-
-    """
-    file_name = os.path.split(results_path)[1]
-    ds_id, version_date_key, stn_id, chunk_id, _, _ = file_name.split('_')
-
-    key_name = tdm.utils.key_patterns[system_version]['results'].format(dataset_id=ds_id, version_date=version_date_key, station_id=stn_id, chunk_id=chunk_id)
-
-    run_date_key = make_run_date_key()
-
-    metadata = {'run_date': run_date_key}
-
-    _ = put_file_s3(s3, bucket, key_name, results_path, metadata, 'application/zstd', retries=5)
-
-    return key_name
-
-
-def put_station(s3, bucket, stn_path, system_version=4):
-    """
-
-    """
-    file_name = os.path.split(stn_path)[1]
-    ds_id, stn_id, _ = file_name.split('_')
-
-    key_name = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=ds_id, station_id=stn_id)
-
-    run_date_key = make_run_date_key()
-
-    metadata = {'run_date': run_date_key}
-
-    _ = put_file_s3(s3, bucket, key_name, stn_path, metadata, 'application/zstd', retries=5)
-
-    return key_name
-
-
-# def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10, threads=50, version=3):
+# def put_result(s3, bucket, results_path, system_version=4):
 #     """
-#     Function to delete Tethys result objects including all object versions.
 
-#     Parameters
-#     ----------
-#     conn_config : dict
-#         A dictionary of the connection info necessary to establish an S3 connection.
-#     bucket : str
-#         The s3 bucket.
-#     dataset_ids : list, str, or None
-#         The specific datasets that should have the results objects removed. None will remove results objects from all datasets in the bucket.
-#     keep_last : int
-#         That last number of runs that should be kept. E.g. a value of 4 will kepp the last 4 runs and remove all prior runs.
-#     threads : int
-#         The number of concurrent threads to use when aggregating the stations and datasets.
-
-#     Returns
-#     -------
-#     list of keys deleted
 #     """
-#     s3 = s3_client(conn_config, threads)
+#     file_name = os.path.split(results_path)[1]
+#     ds_id, version_date_key, stn_id, chunk_id, _, _ = file_name.split('_')
 
-#     if isinstance(dataset_ids, str):
-#         dataset_ids = [dataset_ids]
+#     key_name = tdm.utils.key_patterns[system_version]['results'].format(dataset_id=ds_id, version_date=version_date_key, station_id=stn_id, chunk_id=chunk_id)
 
-#     prefix = tdm.utils.key_patterns[version]['results'].split('{dataset_id}')[0]
+#     run_date_key = make_run_date_key()
 
-#     obj_list = list_object_versions_s3(s3, bucket, prefix)
+#     metadata = {'run_date': run_date_key}
 
-#     obj_list1 = obj_list[obj_list.KeyDate.notnull()].copy()
-#     key_split = obj_list1['Key'].str.split('/')
-#     obj_list1['dataset_id'] = key_split.apply(lambda x: x[2])
-#     obj_list1['station_id'] = key_split.apply(lambda x: x[3])
+#     _ = put_file_s3(s3, bucket, key_name, results_path, metadata, 'application/zstd', retries=5)
 
-#     if isinstance(dataset_ids, list):
-#         obj_list1 = obj_list1[obj_list1['dataset_id'].isin(dataset_ids)].copy()
-
-#     ## Get the keys to the objects that should be removed
-#     obj_list2 = obj_list1.groupby(['dataset_id', 'station_id'])
-
-#     rem_keys = []
-#     for i, row in obj_list2:
-#         rem1 = row.sort_values('KeyDate', ascending=False).iloc[keep_last:]
-#         for i2, row2 in rem1.iterrows():
-#             rem_keys.extend([{'Key': row2['Key'], 'VersionId': row2['VersionId']}])
-
-#     if len(rem_keys) > 0:
-#         ## Split them into 1000 key chunks
-#         rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))
-
-#         ## Run through and delete the objects...
-#         for keys in rem_keys_chunks:
-#             resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})
-
-#     print(str(len(rem_keys)) + ' objects removed')
-#     print('Updating stations and dataset json files...')
-
-#     dataset_list = obj_list1['dataset_id'].unique().tolist()
-
-#     for ds in dataset_list:
-#         ds_stations = put_remote_agg_stations(s3, bucket, ds, threads)
-
-#     ### Aggregate all datasets for the bucket
-#     ds_all = put_remote_agg_datasets(s3, bucket, threads)
-
-#     return rem_keys
+#     return key_name
 
 
-def process_results_versions(dataset_list, bucket, version_data, s3=None, connection_config=None, public_url=None, system_version=4):
+# def put_station(s3, bucket, stn_path, system_version=4):
+#     """
+
+#     """
+#     file_name = os.path.split(stn_path)[1]
+#     ds_id, version_date_key, stn_id, _ = file_name.split('_')
+
+#     key_name = tdm.utils.key_patterns[system_version]['station'].format(dataset_id=ds_id, version_date=version_date_key, station_id=stn_id)
+
+#     run_date_key = make_run_date_key()
+
+#     metadata = {'run_date': run_date_key}
+
+#     _ = put_file_s3(s3, bucket, key_name, stn_path, metadata, 'application/zstd', retries=5)
+
+#     return key_name
+
+
+def delete_dataset_s3(conn_config, bucket, dataset_id, system_version=4):
+    """
+    Function to delete Tethys result objects including all object versions.
+
+    Parameters
+    ----------
+    conn_config : dict
+        A dictionary of the connection info necessary to establish an S3 connection.
+    bucket : str
+        The s3 bucket.
+    dataset_id : str
+        The specific dataset that should be removed.
+
+    Returns
+    -------
+    list of keys deleted
+    """
+    s3 = s3_client(conn_config)
+
+    ## Update/Remove the dataset from the master datasets file
+    dss_key = tdm.utils.key_patterns[system_version]['datasets']
+
+    dss_obj = get_object_s3(dss_key, bucket, s3)
+
+    rem_keys = []
+
+    if dss_obj is not None:
+        dss_list = read_json_zstd(dss_obj)
+
+        dss_list_new = []
+        for dss in dss_list:
+            if dss['dataset_id'] != dataset_id:
+                dss_list_new.append(dss)
+
+        if len(dss_list_new) > 0:
+            content_type = 'application/json'
+
+            run_date_key = make_run_date_key()
+            metadata = {'run_date': run_date_key}
+
+            dss_obj_new = write_json_zstd(dss_list_new)
+
+            resp = put_object_s3(s3, bucket, dss_key, dss_obj_new, metadata, content_type)
+        else:
+            obj_list = list_object_versions_s3(s3, bucket, dss_key)
+            obj_del = obj_list[['Key', 'VersionId']].to_dict('records')
+            resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': obj_del, 'Quiet': True})
+            rem_keys.extend(obj_del)
+
+        ## Remove all of the other objects
+        del_results = tdm.utils.key_patterns[system_version]['results'].split('{version_date}')[0].format(dataset_id=dataset_id)
+        del_versions = tdm.utils.key_patterns[system_version]['versions'].format(dataset_id=dataset_id)
+        del_dataset = tdm.utils.key_patterns[system_version]['dataset'].format(dataset_id=dataset_id)
+
+        del_list = [del_dataset, del_versions, del_results]
+
+        for prefix in del_list:
+
+            obj_list = list_object_versions_s3(s3, bucket, prefix)
+
+            if not obj_list.empty:
+                obj_del = obj_list[['Key', 'VersionId']].to_dict('records')
+
+                ## Split them into 1000 key chunks
+                rem_keys_chunks = np.array_split(obj_del, int(np.ceil(len(obj_del)/1000)))
+
+                ## Run through and delete the objects...
+                for keys in rem_keys_chunks:
+                    resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})
+
+                rem_keys.extend(obj_del)
+    else:
+        print('No dataset found with the given dataset_id.')
+
+    return rem_keys
+
+
+def process_dataset_versions(dataset_list, bucket, version_data, s3=None, connection_config=None, public_url=None, system_version=4):
     """
     Function to process the run date keys for all datasets for the extraction.
 
     Parameters
     ----------
-    processing_code : int
-        The processing code to determine how the input data should be processed.
     dataset_list : list of dict
         The list of datasets, which is the output of the process_datasets function.
     remote : dict
@@ -654,9 +670,6 @@ def process_results_versions(dataset_list, bucket, version_data, s3=None, connec
             A dictionary of the connection info necessary to establish an S3 connection.
         bucket : str
             The S3 bucket.
-    run_date : str, datetime, date, pd.Timestamp, or None
-        The run_date to use for processing. If None, then the run_date will be generated.
-
 
     Returns
     -------
@@ -684,6 +697,8 @@ def process_results_versions(dataset_list, bucket, version_data, s3=None, connec
     t1 = Tethys([remote1])
     datasets = t1._datasets
 
+    old_versions = {}
+
     for ds_id, vd in version_dict.items():
         if ds_id in datasets:
             new_date = pd.Timestamp(vd['version_date'])
@@ -691,8 +706,9 @@ def process_results_versions(dataset_list, bucket, version_data, s3=None, connec
             old_date = pd.to_datetime([pd.Timestamp(v['version_date']) for v in vd_old]).max()
             if new_date < old_date:
                 raise ValueError('The new version date is prior to the max old version date. New version date: {new_date}; old max version date: {old_date}. dataset_id: {ds_id}.'.format(new_date=new_date.isoformat(), old_date=old_date.isoformat(), ds_id=ds_id))
+            old_versions[ds_id] = vd_old
 
-    return version_dict
+    return version_dict, old_versions
 
 
 # def reprocess_datasets(dataset, conn_config, bucket, public_url, threads=20):
@@ -777,7 +793,7 @@ def process_results_versions(dataset_list, bucket, version_data, s3=None, connec
 #     print('Finished')
 
 
-def determine_results_chunks_diffs(source_paths, remote):
+def determine_results_chunks_diffs(source_paths, remote, add_old=False):
     """
 
     """
@@ -797,8 +813,8 @@ def determine_results_chunks_diffs(source_paths, remote):
         # print(ds_id)
 
         if ds_id in datasets:
-            dataset = datasets[ds_id]
-            method = dataset['method']
+            # dataset = datasets[ds_id]
+            # method = dataset['method']
             vd_old = tethys.get_versions(ds_id)
             vd_max = g.version_date.max()
 
@@ -812,40 +828,28 @@ def determine_results_chunks_diffs(source_paths, remote):
             vd1 = [vd['version_date'] for vd in vd_old if pd.Timestamp(vd['version_date']) == vd_max]
 
             if len(vd1) > 0:
-                stns = tethys.get_stations(ds_id)
-                old_stn_ids = [s['station_id'] for s in stns]
-
-                vd = vd1[0]
-                station_ids = g.station_id.tolist()
+                vd = vd1[-1]
                 merge_vals = ['conflict', 'new']
             else:
-                if method in ['sensor_recording', 'field_activity', 'sample_analysis']:
-                    stns = tethys.get_stations(ds_id)
-                    old_stn_ids = [s['station_id'] for s in stns]
-
+                if add_old:
                     vd = vd_old[-1]['version_date']
-                    station_ids = old_stn_ids
                     merge_vals = ['conflict', 'new', 'old']
                 else:
-                    # vd = vd_max
-                    # station_ids = g.station_id.tolist()
-                    # merge_vals = ['conflict', 'new']
                     g['key'] = np.nan
                     g['_merge'] = 'new'
                     diffs_list.append(g)
                     continue
 
-            # Determine which stns already exist in Tethys
-            old_stns = list(set([s for s in station_ids if s in old_stn_ids]))
+            # Load in the results chunks
+            vd_key = make_run_date_key(vd)
+            rc_key = tdm.utils.key_patterns[4]['results_chunks'].format(dataset_id=ds_id, version_date=vd_key)
 
-            # Load in the appropriate chunks to check
-            tethys._load_results_chunks(ds_id, old_stns)
+            s_remote = copy.deepcopy(remote)
+            _ = s_remote.pop('version')
+            s_remote['obj_key'] = rc_key
 
-            chunks1 = []
-            for stn_id, c in tethys._results_chunks[ds_id].items():
-                # print(stn_id)
-                chunks2 = chunk_filters(c, version_date=vd)
-                chunks1.extend(chunks2)
+            rc_obj = get_object_s3(**s_remote)
+            chunks1 = read_json_zstd(rc_obj)
 
             chunks_df = pd.DataFrame(chunks1).drop(['version_date', 'content_length', 'height', 'chunk_day'], axis=1).rename(columns={'chunk_hash': 'original_chunk_hash'})
 
@@ -855,10 +859,8 @@ def determine_results_chunks_diffs(source_paths, remote):
             combo1 = combo1[combo1['_merge'].isin(merge_vals)].copy()
             combo1['version_date'] = vd_max
 
-            if 'old' in merge_vals:
-                combo1.loc[combo1['chunk_hash'] == combo1['original_chunk_hash'], '_merge'] = 'new'
-            else:
-                combo1 = combo1[combo1['chunk_hash'] != combo1['original_chunk_hash']].copy()
+            combo1['_merge'] = combo1['_merge'].cat.set_categories(['identical', 'conflict', 'new', 'old'])
+            combo1.loc[combo1['chunk_hash'] == combo1['original_chunk_hash'], '_merge'] = 'identical'
 
             diffs_list.append(combo1.drop(['original_chunk_hash'], axis=1))
         else:
@@ -871,8 +873,64 @@ def determine_results_chunks_diffs(source_paths, remote):
     return combo2
 
 
+# def update_latest_stations(dataset_id, bucket, s3, public_url=None, threads=60, system_version=4):
+#     """
 
+#     """
+#     ## Get the versions
+#     rv_key = tdm.utils.key_patterns[system_version]['versions'].format(dataset_id=dataset_id)
 
+#     v_obj1 = get_object_s3(rv_key, bucket, s3=s3, public_url=public_url, counter=2)
+
+#     if v_obj1 is not None:
+#         v2 = read_json_zstd(v_obj1)
+
+#         last_version_date = max([v['version_date'] for v in v2])
+
+#         ## Update all of the individual station objects
+#         # stn_key_str = tdm.utils.key_patterns[system_version]['station']
+#         # vd = make_run_date_key(version_date)
+#         # stn_prefix = stn_key_str.split('{station_id}')[0].format(dataset_id=dataset_id, verion_date=vd)
+#         #
+#         # if system_version >= 4:
+#         #     list1 = list_objects_s3(s3, bucket, stn_prefix, delimiter='/')
+#         # else:
+#         #     raise ValueError('Wrong object structure version.')
+#         #
+#         # list2 = list1[list1.Key.str.contains('station.json.zst')].copy()
+#         #
+#         # source_dest_keys = []
+#         # for key in list2['Key'].tolist():
+#         #     stn_id = key.split('/')[4].split('.')[0]
+#         #     stn_key = stn_key_str.format(dataset_id=dataset_id, verion_date=vd, station_id=stn_id)
+#         #     dict1 = {'source_key': key, 'dest_key': stn_key}
+#         #     source_dest_keys.append(dict1)
+#         #
+#         # resp_list = multi_copy_object_s3(s3, bucket, bucket, source_dest_keys, retries=5, threads=threads)
+
+#         ## Update the stations object
+#         stns_key_str = tdm.utils.key_patterns[system_version]['stations']
+#         vd = make_run_date_key(last_version_date)
+#         stns_key = stns_key_str.format(dataset_id=dataset_id, verion_date=vd)
+
+#         dest_key_str = tdm.utils.key_patterns[system_version]['latest_stations']
+#         dest_key = dest_key_str.format(dataset_id=dataset_id)
+
+#         stn_resp = copy_object_s3(s3, bucket, bucket, stns_key, dest_key)
+
+#         ## Update the results_chunks object
+#         rc_key_str = tdm.utils.key_patterns[system_version]['results_chunks']
+#         vd = make_run_date_key(last_version_date)
+#         rc_key = rc_key_str.format(dataset_id=dataset_id, verion_date=vd)
+
+#         dest_key_str = tdm.utils.key_patterns[system_version]['latest_results_chunks']
+#         dest_key = dest_key_str.format(dataset_id=dataset_id)
+
+#         rc_resp = copy_object_s3(s3, bucket, bucket, rc_key, dest_key)
+
+#         return True
+#     else:
+#         return False
 
 
 
