@@ -13,8 +13,6 @@ import tethys_utils as tu
 from tethysts.utils import s3_client, get_object_s3, read_json_zstd, read_pkl_zstd, download_results, create_public_s3_url
 import tethys_data_models as tdm
 from typing import List, Optional, Dict, Union
-# import boto3
-# import botocore
 from botocore import exceptions as bc_exceptions
 import requests
 import uuid
@@ -25,15 +23,13 @@ import concurrent.futures
 import copy
 from shapely import wkb, wkt, geometry
 import pathlib
-import stat
 import dask
-from hashlib import blake2b
 import glob
 from datetime import date, datetime
 import multiprocessing as mp
 import pyproj
 import random
-from tethysts import Tethys
+from time import sleep
 
 ##############################################
 ### Parameters
@@ -65,23 +61,23 @@ tz_str = 'Etc/GMT{0:+}'
 
 
 
-def save_chunks(chunks_list, output_path):
-    """
+# def save_chunks(chunks_list, output_path):
+#     """
 
-    """
-    for c in chunks_list:
-        chunk_json = tdm.dataset.ChunkID(height=int(c.height.values[0] * 1000), start_date=int(c.attrs['start_date'])).json(exclude_none=True).encode('utf-8')
-        chunk_id = blake2b(chunk_json, digest_size=12).hexdigest()
-        stn_id = str(c.station_id.values[0])
-        file_path = os.path.join(output_path, stn_id + '_' + chunk_id + '.nc.zst')
+#     """
+#     for c in chunks_list:
+#         chunk_json = tdm.dataset.ChunkID(height=int(c.height.values[0] * 1000), start_date=int(c.attrs['start_date'])).json(exclude_none=True).encode('utf-8')
+#         chunk_id = blake2b(chunk_json, digest_size=12).hexdigest()
+#         stn_id = str(c.station_id.values[0])
+#         file_path = os.path.join(output_path, stn_id + '_' + chunk_id + '.nc.zst')
 
-        b = c.copy().load()
+#         b = c.copy().load()
 
-        obj1 = tu.misc.write_pkl_zstd(b.to_netcdf(), file_path)
+#         obj1 = tu.misc.write_pkl_zstd(b.to_netcdf(), file_path)
 
-        b.close()
-        del b
-        del obj1
+#         b.close()
+#         del b
+#         del obj1
 
 
 def multi_save_dataset_stations(nc_paths, block_length, block_length_factor: int = 10, compression='zstd', remove_station_data=True, max_workers=3):
@@ -582,7 +578,7 @@ class Titan(object):
 
     """
 
-    def __init__(self, temp_path: pathlib.Path, run_id: str = None, key: str = None, diagnostics_url: HttpUrl = None ):
+    def __init__(self, temp_path: pathlib.Path, add_old: bool, run_id: str = None, key: str = None, diagnostics_url: HttpUrl = None ):
         """
 
         """
@@ -604,6 +600,8 @@ class Titan(object):
         self.previous_path = previous_path
         # self.final_path = final_path
 
+        self._add_old = add_old
+
         ## Test for the run_id and key
         need_dgn = True
 
@@ -640,6 +638,7 @@ class Titan(object):
         self.key = key
         self.run_id = run_id
         self.temp_path = temp_path
+        self._results_objects_updated = []
 
         ## Set up diagnostics dict
         if need_dgn:
@@ -736,15 +735,15 @@ class Titan(object):
         # main_vars = [parameter] + ancillary_variables
 
         # Convert times to local TZ if necessary
-        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+        if (not freq_code in ['None', 'T', 'H', '1H']) and (utc_offset != '0H'):
             t1 = int(utc_offset.split('H')[0])
             tz1 = tz_str.format(-t1)
             df1['time'] = df1['time'].dt.tz_localize('UTC').dt.tz_convert(tz1).dt.tz_localize(None)
 
         ## Aggregate data if necessary
         # Parameter
-        if freq_code == 'T':
-            data1 = df1
+        if freq_code == 'None':
+            data1 = df1.drop_duplicates(subset=['time']).sort_values('time')
         else:
             agg_fun = agg_stat_mapping[agg_stat]
 
@@ -765,25 +764,15 @@ class Titan(object):
                     av_list.append(av1)
 
             # Put the data together
-            data1 = pd.concat(av_list, axis=1).reset_index()
+            data1 = pd.concat(av_list, axis=1).reset_index().sort_values('time')
 
         # Convert time back to UTC if necessary
-        if (not freq_code in ['T', 'H', '1H']) and (not utc_offset == '0H'):
+        if (not freq_code in ['None', 'T', 'H', '1H']) and (utc_offset != '0H'):
             data1['time'] = data1['time'].dt.tz_localize(tz1).dt.tz_convert('utc').dt.tz_localize(None)
 
         return data1
 
 
-
-
-    # def init_results_dict(self):
-    #     """
-
-    #     """
-    #     results_dict = {d: [] for d in self.datasets}
-    #     setattr(self, 'results_dict', results_dict)
-
-
     def status_checks(self, connection_config: tdm.base.ConnectionConfig, bucket: str, public_url: str = None, system_version: int = 4, load_diagnostics: bool = False):
         """
 
@@ -966,16 +955,17 @@ class Titan(object):
         """
         tu.misc.diagnostic_check(self.diagnostics, 'load_dataset_metadata')
 
-        version_dict = tu.s3.process_results_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
-        max_version_date_key = tu.misc.make_run_date_key(max(list([d['version_date'] for d in version_dict.values()])))
+        version_dict, vd_old = tu.s3.process_dataset_versions(self.dataset_list, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, version_data=version_data, system_version=self.system_version)
+        max_version_date = max(list([d['version_date'] for d in version_dict.values()]))
 
         # setattr(self, 'processing_code', processing_code)
         setattr(self, 'version_dict', version_dict)
-        setattr(self, 'max_version_date_key', max_version_date_key)
+        setattr(self, 'max_version_date', max_version_date)
+        setattr(self, 'old_versions', vd_old)
 
         ## diagnostic log
         self.diagnostics['process_versions'] = {'pass': True}
-        self.diagnostics['attributes'].update({'version_dict': version_dict, 'max_version_date_key': max_version_date_key})
+        self.diagnostics['attributes'].update({'version_dict': version_dict, 'max_version_date': max_version_date, 'old_versions': vd_old})
 
         return version_dict
 
@@ -1014,12 +1004,12 @@ class Titan(object):
             # print(str(d.ref.values[0]))
 
             ## Update the metadata
-            d = tu.processing.add_metadata_results(d, dataset1, self.max_version_date_key)
+            d = tu.processing.add_metadata_results(d, dataset1, self.max_version_date)
 
             file_id = uuid.uuid4().hex[:14]
             file_path = os.path.join(self.preprocessed_path, preprocessed_file_str.format(ds_id=dataset_id, file_id=file_id))
-            # d.to_netcdf(file_path)
-            _ = tu.misc.write_pkl_zstd(d.to_netcdf(), file_path)
+
+            tu.processing.write_nc_zstd(d, file_path)
 
             # Read the file back in to make sure it worked...
             # _ = xr.load_dataset(read_pkl_zstd(file_path))
@@ -1098,7 +1088,7 @@ class Titan(object):
         """
 
         """
-        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_list, self.max_version_date_key, func_dict)
+        new_paths2 = multi_calc_new_variables(source_paths, self.dataset_list, self.max_version_date, func_dict)
 
         return new_paths2
 
@@ -1229,7 +1219,7 @@ class Titan(object):
         tu.misc.diagnostic_check(self.diagnostics, 'process_versions')
 
         ## Put the big files first
-        merge_list = [[p, os.path.getsize(p)] for p in source_paths]
+        merge_list = [[p, os.path.getsize(p)] for p in list(set(source_paths))]
         merge_df = pd.DataFrame(merge_list, columns=['path', 'size'])
 
         source_paths1 = merge_df.sort_values('size', ascending=False).path.tolist()
@@ -1240,7 +1230,7 @@ class Titan(object):
             for source_path in source_paths1:
                 ds_id = source_path.split('/')[-1].split('_')[0]
                 metadata = self.datasets[ds_id]
-                f = tu.processing.save_new_results(source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
+                f = tu.processing.save_new_results(source_path, metadata, self.max_version_date, correct_times, self.system_version)
                 results_new_paths1.extend(f)
         else:
             with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
@@ -1248,7 +1238,7 @@ class Titan(object):
                 for source_path in source_paths1:
                     ds_id = source_path.split('/')[-1].split('_')[0]
                     metadata = self.datasets[ds_id]
-                    f = executor.submit(tu.processing.save_new_results, source_path, metadata, self.max_version_date_key, correct_times, self.system_version)
+                    f = executor.submit(tu.processing.save_new_results, source_path, metadata, self.max_version_date, correct_times, self.system_version)
                     futures.append(f)
                 runs = concurrent.futures.wait(futures)
 
@@ -1265,19 +1255,34 @@ class Titan(object):
         return results_new_paths1
 
 
+    def _determine_results_chunks_diffs(self, source_paths, remote, add_old=False):
+        """
+
+        """
+        r_chunks = tu.s3.determine_results_chunks_diffs(source_paths, remote, add_old)
+
+        self._results_chunks_diffs = r_chunks
+
+        return r_chunks
+
+
     def update_conflicting_results(self, results_paths, threads=60, max_workers=4):
         """
 
         """
+        tu.misc.diagnostic_check(self.diagnostics, 'save_new_results')
+
         ## Determine which chunks need to uploaded
         if isinstance(self.public_url, str):
             remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
         else:
             remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
 
-        r_chunks = tu.s3.determine_results_chunks_diffs(results_paths, remote1)
+        results_paths_set = set(results_paths)
 
-        r_chunks1 = r_chunks[r_chunks['_merge'].isin(['old', 'conflict'])].copy()
+        r_chunks = self._determine_results_chunks_diffs(list(results_paths_set), remote1, add_old=self._add_old)
+
+        r_chunks1 = r_chunks[r_chunks['_merge'].isin(['conflict'])].copy()
 
         if not r_chunks1.empty:
             chunks = r_chunks1.to_dict('records')
@@ -1301,102 +1306,176 @@ class Titan(object):
             if max_workers <= 1:
                 updated_paths = []
                 for chunk in chunks1:
-                    p1 = tu.processing.update_compare_results(chunk, self.max_version_date_key, self.preprocessed_path)
+                    ds_id = str(chunk).split('/')[-3]
+                    metadata = self.datasets[ds_id]
+                    p1 = tu.processing.update_compare_results(chunk, metadata, self.max_version_date, self.preprocessed_path)
                     updated_paths.append(p1)
             else:
                 with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
                     futures = []
                     for chunk in chunks1:
-                        f = executor.submit(tu.processing.update_compare_results, chunk, self.max_version_date_key, self.preprocessed_path)
+                        ds_id = str(chunk).split('/')[-3]
+                        metadata = self.datasets[ds_id]
+                        f = executor.submit(tu.processing.update_compare_results, chunk, metadata, self.max_version_date, self.preprocessed_path)
                         futures.append(f)
                     runs = concurrent.futures.wait(futures)
 
                 updated_paths = [r.result() for r in runs[0]]
+
+            ## Remove previous results
+            for chunk in chunks1:
+                if chunk.exists():
+                    os.remove(chunk)
+
         else:
             updated_paths = []
 
-        ## Combine paths with original paths
-        combo_paths = list(set(results_paths).union(set(updated_paths)))
+        ## Combine updated paths with "new" paths
+        new_data_paths = r_chunks[r_chunks['_merge'] == 'new']['file_path'].tolist()
+        updated_paths.extend(new_data_paths)
+        updated_paths.sort()
+
+        ## Remove files labeled as "identical"
+        ident1 = r_chunks[r_chunks['_merge'] == 'identical']
+
+        if not ident1.empty:
+            paths = ident1['file_path'].tolist()
+            for path in paths:
+                if os.path.exists(path):
+                    os.remove(path)
 
         ## diagnostic log
         self.diagnostics['update_conflicting_results'] = {'pass': True}
 
-        return combo_paths
+        return updated_paths
 
 
-    def update_stations(self, results_paths, threads=60, max_workers=4):
+    def _update_results_chunks(self, results_paths, max_workers=4):
         """
 
         """
-        ## Organise the paths
-        paths_dict = {}
-        for path in results_paths:
-            path1 = os.path.split(path)[1]
-            ds_id, version_date1, stn_id, chunk_id, _, _ = path1.split('_')
+        ## checks
+        tu.misc.diagnostic_check(self.diagnostics, 'save_new_results')
 
-            if ds_id in paths_dict:
-                if stn_id in paths_dict[ds_id]:
-                    paths_dict[ds_id][stn_id]['results_paths'].append(path)
-                else:
-                    paths_dict[ds_id].update({stn_id: {'results_paths': [path], 'new_path': self.preprocessed_path}})
-            else:
-                paths_dict[ds_id] = {stn_id: {'results_paths': [path], 'new_path': self.preprocessed_path}}
+        add_old = self._add_old
+        if self._add_old:
+            tu.misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
 
-        ## Get the old stations
         if isinstance(self.public_url, str):
             remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
         else:
             remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
 
-        t1 = Tethys([remote1])
+        results_paths_set = set(results_paths)
+        r_chunks = self._determine_results_chunks_diffs(list(results_paths_set), remote1)
+        r_chunks1 = r_chunks[r_chunks['_merge'] == 'conflict']
 
-        dataset_ids = [ds_id for ds_id in paths_dict]
+        if not r_chunks1.empty:
+            tu.misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
 
-        for ds_id in dataset_ids:
-            if ds_id in t1._datasets:
-                _ = t1.get_stations(ds_id)
+        ## Organise the paths
+        paths_dict = {}
+        for path in list(set(results_paths)):
+            path1 = os.path.split(path)[1]
+            ds_id, version_date1, stn_id, chunk_id, _, _ = path1.split('_')
 
-        old_stn_dict = t1._stations
+            if ds_id in paths_dict:
+                    paths_dict[ds_id].append(path)
+            else:
+                paths_dict[ds_id] = [path]
 
-        if len(old_stn_dict) > 0:
-            remote1['system_version'] = remote1.pop('version')
+        ## Iterate through datasets
+        futures = []
 
-            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
-                futures = []
-                for ds_id, stn_path in paths_dict.items():
-                    if ds_id in old_stn_dict:
-                        for stn_id in stn_path:
-                            if stn_id in old_stn_dict[ds_id]:
-                                remote1['dataset_id'] = ds_id
-                                remote1['station_id'] = stn_id
-                                f = executor.submit(tu.s3.get_remote_station, **remote1)
-                                futures.append(f)
-                runs = concurrent.futures.wait(futures)
+        for ds_id, stn_files in paths_dict.items():
 
-            old_stns = [r.result() for r in runs[0]]
+            if ds_id in self.old_versions:
+                old_versions = [v['version_date'] for v in self.old_versions[ds_id]]
+            else:
+                old_versions = []
 
-            for stn in old_stns:
-                ds_id = stn['dataset_id']
-                stn_id = stn['station_id']
-                paths_dict[ds_id][stn_id]['old_stn_data'] = stn
+            ## Get the old results chunks file
+            if self.max_version_date in old_versions:
+                old_rc_data = tu.s3.get_remote_results_chunks(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=self.max_version_date, system_version=self.system_version)
+            elif add_old and (len(old_versions) > 0):
+                previous_version_date = old_versions[-1]
+                old_rc_data = tu.s3.get_remote_results_chunks(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=previous_version_date, system_version=self.system_version)
+            else:
+                old_rc_data = None
 
-        ## Update station data
-        if max_workers <= 1:
-            stn_paths = []
-            for ds_id, stn in paths_dict.items():
-                for stn_id, d in stn.items():
-                    f = tu.processing.update_station(**d)
-                    stn_paths.append(f)
-        else:
             with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
-                futures = []
-                for ds_id, stn in paths_dict.items():
-                    for stn_id, d in stn.items():
-                        f = executor.submit(tu.processing.update_station, **d)
-                        futures.append(f)
-                runs = concurrent.futures.wait(futures)
+                f = executor.submit(tu.processing.update_results_chunks, stn_files, self.preprocessed_path, old_rc_data)
+                futures.append(f)
 
-            stn_paths = [r.result() for r in runs[0]]
+        runs = concurrent.futures.wait(futures)
+
+        rc_paths = [r.result() for r in runs[0]]
+
+        self.results_chunks_paths = rc_paths
+
+        ## diagnostic log
+        self.diagnostics['update_results_chunks'] = {'pass': True}
+
+        return rc_paths
+
+
+    def _update_stations(self, results_paths, max_workers=4):
+        """
+
+        """
+        ## Checks
+        tu.misc.diagnostic_check(self.diagnostics, 'update_results_chunks')
+
+        add_old = self._add_old
+
+        ## Organise the paths
+        paths_dict = {}
+        for path in list(set(results_paths)):
+            path1 = os.path.split(path)[1]
+            ds_id, version_date1, stn_id, chunk_id, _, _ = path1.split('_')
+
+            if ds_id in paths_dict:
+                    paths_dict[ds_id].append(path)
+            else:
+                paths_dict[ds_id] = [path]
+
+        rc_paths = self.results_chunks_paths.copy()
+
+        rc_path_dict = {}
+        for path in rc_paths:
+            path1 = os.path.split(path)[1]
+            ds_id = path1.split('_')[0]
+            rc_path_dict[ds_id] = path
+
+        ## Iterate through datasets
+        futures = []
+
+        for ds_id, data_paths in paths_dict.items():
+            rc_path = rc_path_dict[ds_id]
+
+            if ds_id in self.old_versions:
+                old_versions = [v['version_date'] for v in self.old_versions[ds_id]]
+            else:
+                old_versions = []
+
+            ## Get the old results chunks file
+            if self.max_version_date in old_versions:
+                old_stns_data = tu.s3.get_remote_stations(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=self.max_version_date, system_version=self.system_version)
+            elif add_old and (len(old_versions) > 0):
+                previous_version_date = old_versions[-1]
+                old_stns_data = tu.s3.get_remote_stations(self.bucket, s3=None, connection_config=self.connection_config, public_url=self.public_url, dataset_id=ds_id, version_date=previous_version_date, system_version=self.system_version)
+            else:
+                old_stns_data = None
+
+            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+                f = executor.submit(tu.processing.update_stations, data_paths, rc_path, self.preprocessed_path, old_stns_data)
+                futures.append(f)
+
+        runs = concurrent.futures.wait(futures)
+
+        stn_paths = [r.result() for r in runs[0]]
+
+        self.stations_paths = stn_paths
 
         ## diagnostic log
         self.diagnostics['update_stations'] = {'pass': True}
@@ -1404,140 +1483,173 @@ class Titan(object):
         return stn_paths
 
 
-    def upload_final_results(self, results_paths, threads=60):
+    def _update_versions(self, max_workers=4):
         """
 
         """
-        tu.misc.diagnostic_check(self.diagnostics, 'save_new_results')
+        ## Checks
         tu.misc.diagnostic_check(self.diagnostics, 'update_stations')
 
-        ## Determine which chunks need to uploaded
-        if isinstance(self.public_url, str):
-            remote1 = {'bucket': self.bucket, 'public_url': self.public_url, 'version': self.system_version}
-        else:
-            remote1 = {'bucket': self.bucket, 'connection_config': self.connection_config, 'version': self.system_version}
+        ## Determine what datasets are updated
+        dataset_ids = []
+        for d in self.results_chunks_paths:
+            ds_id = os.path.split(d)[1].split('_')[0]
+            dataset_ids.append(ds_id)
 
-        r_chunks = tu.s3.determine_results_chunks_diffs(results_paths, remote1)
+        ## Run versions update
+        futures = []
+        for dataset_id in dataset_ids:
+            version_data1 = self.version_dict[dataset_id]
 
-        stn_keys = []
-        results_keys = []
+            ## Get the versions
+            if dataset_id in self.old_versions:
+                rv_key = tdm.utils.key_patterns[self.system_version]['versions'].format(dataset_id=dataset_id)
 
-        if not r_chunks.empty:
-            # old_chunks = r_chunks[r_chunks['_merge'] == 'old']
+                v_obj1 = get_object_s3(rv_key, self.bucket, connection_config=self.connection_config, public_url=self.public_url)
 
-            # if not old_chunks.empty:
-            #     ## Direct copy S3 objects
-            #     print('Old objects will be copied over to the new version.')
-
-            #     s3 = s3_client(self.connection_config, max_pool_connections=30)
-
-            #     old_chunks_list = old_chunks.to_dict('records')
-
-            #     source_dest_keys = []
-            #     for chunk in old_chunks_list:
-            #         version_date_key = tu.misc.make_run_date_key(chunk['version_date'])
-            #         dest_key = tdm.utils.key_patterns[self.system_version]['results'].format(dataset_id=chunk['dataset_id'], version_date=version_date_key, station_id=chunk['station_id'], chunk_id=chunk['chunk_id'])
-
-            #         if dest_key != chunk['key']:
-            #             source_dest_keys.append({'source_key': chunk['key'], 'dest_key': dest_key})
-
-            #     if len(source_dest_keys) > 0:
-            #         _ = tu.s3.multi_copy_object_s3(s3, self.bucket, self.bucket, source_dest_keys)
-
-            if 'old' in r_chunks._merge.unique():
-                raise ValueError('Looks like you need to run the update_conflicting_results method.')
-
-            other_chunks = r_chunks[r_chunks['_merge'].isin(['conflict', 'new'])]
-
-            if not other_chunks.empty:
-
-                merge_types = other_chunks['_merge'].unique()
-
-                if 'conflict' in merge_types:
-                    tu.misc.diagnostic_check(self.diagnostics, 'update_conflicting_results')
-
-                results_paths1 = other_chunks['file_path'].tolist()
-
-                random.shuffle(results_paths1)
-
-                ## Make sure all of the station files exist
-                stn_paths = []
-                for path in results_paths1:
-                    ds_id, version_date1, stn_id, chunk_id, _, _ = os.path.split(path)[1].split('_')
-
-                    stn_file_name = tu.processing.ds_stn_json_str.format(ds_id=ds_id, stn_id=stn_id)
-                    stn_file_path = os.path.join(os.path.split(path)[0], stn_file_name)
-
-                    if os.path.exists(stn_file_path):
-                        stn_paths.append(stn_file_path)
-                    else:
-                        raise FileNotFoundError('station file not found: ' + stn_file_path)
-
-                ## Iterate through the results and stations
-                with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
-                    s3 = s3_client(self.connection_config, max_pool_connections=threads)
-                    futures = []
-
-                    for path in results_paths1:
-                        f = executor.submit(tu.s3.put_result, s3, self.bucket, path, system_version=self.system_version)
-                        futures.append(f)
-
-                    for path in set(stn_paths):
-                        # Add the station file
-                        f = executor.submit(tu.s3.put_station, s3, self.bucket, path, system_version=self.system_version)
-                        futures.append(f)
-
-                    runs = concurrent.futures.wait(futures)
-
-                keys = [r.result() for r in runs[0]]
-                keys.sort()
-
-                stn_keys = [key for key in keys if 'station' in key]
-                results_keys = [key for key in keys if 'results' in key]
+                old_versions = read_json_zstd(v_obj1)
             else:
-                print('No chunks have changed or have been added, so no files have been uploaded.')
+                old_versions = None
+
+            with concurrent.futures.ProcessPoolExecutor(max_workers=max_workers, mp_context=mp.get_context("spawn")) as executor:
+                f = executor.submit(tu.processing.update_versions, version_data1, self.preprocessed_path, old_versions)
+                futures.append(f)
+
+        runs = concurrent.futures.wait(futures)
+
+        versions_paths = [r.result() for r in runs[0]]
+
+        self.versions_paths = versions_paths
+
+        ## diagnostic log
+        self.diagnostics['update_versions'] = {'pass': True}
+
+        return versions_paths
+
+
+    def _update_datasets(self):
+        """
+
+        """
+        ## Checks
+        tu.misc.diagnostic_check(self.diagnostics, 'update_versions')
+
+        ## Save the individual datasets
+        ds_paths = []
+        for stns_path in self.stations_paths:
+            ds_id = os.path.split(stns_path)[1].split('_')[0]
+            dataset = self.datasets[ds_id]
+            ds_path = tu.processing.update_dataset(dataset, self.preprocessed_path, stns_path, self.system_version)
+            ds_paths.append(ds_path)
+
+        ## Process the agg dataset file
+        # Get the old file
+        dss_key = tdm.utils.key_patterns[self.system_version]['datasets']
+
+        dss_obj = get_object_s3(dss_key, bucket=self.bucket, connection_config=self.connection_config, public_url=self.public_url, counter=2)
+
+        if dss_obj is None:
+            old_datasets = None
         else:
-            print('No chunks have changed or have been added, so no files have been uploaded.')
+            old_datasets = read_json_zstd(dss_obj)
 
-        ## Remove source files
-        path1 = pathlib.Path(os.path.split(results_paths[0])[0])
-        for path in path1.glob('*_results.nc.zst'):
-            os.remove(path)
-        for path in path1.glob('*_station.json.zst'):
-            os.remove(path)
+        # Process and save the new file
+        dss_path = tu.processing.update_dataset_agg(ds_paths, self.preprocessed_path, old_datasets)
 
-        return results_keys, stn_keys
+        ## diagnostic log
+        self.diagnostics['update_datasets'] = {'pass': True}
+
+        return ds_paths
 
 
-    def update_aggregates(self, threads=60):
+    def update_auxiliary_objects(self, results_paths, max_workers=4):
         """
 
         """
-        ## Update the datasets and station jsons
-        s3 = s3_client(self.connection_config, threads)
+        if len(results_paths) > 0:
+            print('-- updating the results chunks...')
+            rc_paths = self._update_results_chunks(results_paths, max_workers)
 
-        for ds in self.dataset_list:
-            dataset_id = ds['dataset_id']
+            print('-- updating the stations...')
+            stns_paths = self._update_stations(results_paths, max_workers)
 
-            print(dataset_id)
+            print('-- updating the versions...')
+            versions_paths = self._update_versions(max_workers)
 
-            print('-- Aggregating stations')
-            stns_list = tu.s3.put_remote_agg_stations(dataset_id, self.bucket, s3, self.public_url, threads=threads, system_version=self.system_version)
+            print('-- updating the datasets...')
+            ds_paths = self._update_datasets()
+        else:
+            print('No results to update')
 
-            if stns_list:
-                # print('-- Updating versions')
-                version_data1 = self.version_dict[dataset_id]
 
-                version_list =  tu.s3.update_versions(version_data1, self.bucket, self.connection_config, self.public_url, system_version=self.system_version)
+    def upload_final_objects(self, threads=60):
+        """
 
-                # print('-- Updating Dataset')
-                ds_new = tu.s3.put_dataset(ds, self.bucket, s3=s3, public_url=self.public_url, system_version=self.system_version)
+        """
+        ## Determine all the different files to upload
+        rc_paths = glob.glob(os.path.join(self.preprocessed_path, '*results_chunks.json.zst'))
 
-        # Aggregate all datasets for the bucket
-        print('-- Aggregating datasets')
-        ds_all = tu.s3.put_remote_agg_datasets(self.bucket, s3, self.public_url, 30, system_version=self.system_version)
+        if len(rc_paths) > 0:
 
-        return ds_all
+            results_paths = glob.glob(os.path.join(self.preprocessed_path, '*results.nc.zst'))
+            stns_paths = glob.glob(os.path.join(self.preprocessed_path, '*stations.json.zst'))
+            v_paths = glob.glob(os.path.join(self.preprocessed_path, '*versions.json.zst'))
+            ds_paths = glob.glob(os.path.join(self.preprocessed_path, '*dataset.json.zst'))
+            dss_path = glob.glob(os.path.join(self.preprocessed_path, 'datasets.json.zst'))
+
+            ## Checks
+            tu.misc.diagnostic_check(self.diagnostics, 'update_datasets')
+
+            ds_set = set()
+            for p in results_paths:
+                ds_id = os.path.split(p)[1].split('_')[0]
+                ds_set.add(ds_id)
+
+            if not (len(rc_paths) == len(stns_paths) == len(ds_paths) == len(ds_set) == len(v_paths)):
+                raise ValueError('The number of auxillary files do not match up.')
+
+            if len(dss_path) != 1:
+                raise ValueError('There must be one datasets.json.zst file.')
+
+            ## Prepare files to be uploaded
+            all_paths = results_paths.copy()
+            all_paths.extend(rc_paths)
+            all_paths.extend(stns_paths)
+            all_paths.extend(v_paths)
+            all_paths.extend(ds_paths)
+            all_paths.extend(dss_path)
+            random.shuffle(all_paths)
+
+            s3 = s3_client(self.connection_config, max_pool_connections=threads)
+
+            s3_list = []
+            append = s3_list.append
+            for path in all_paths:
+                dict1 = tu.processing.prepare_file_for_s3(path, s3, self.bucket, self.system_version)
+                append(dict1)
+
+            ## Iterate through the files
+            with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
+                futures = []
+
+                for s3_dict in s3_list:
+                    f = executor.submit(tu.s3.put_file_s3, **s3_dict)
+                    futures.append(f)
+
+                runs = concurrent.futures.wait(futures)
+
+            resp = [r.result() for r in runs[0]]
+
+        else:
+            print('There seems to be no useful files in the results path. Nothing will get uploaded and any files in the path will be removed.')
+            resp = None
+
+        ## Remove all files in the results path
+        path1 = pathlib.Path(self.preprocessed_path)
+        for path in path1.glob('*'):
+            os.remove(path)
+
+        return resp
 
 
     def clear_interim_results(self):
